class Args:
    """
    This module helps in preventing args being sent through multiple of classes to reach
    any analysis/laser module
    """

    def __init__(self):
        self.solver_timeout = 10000
        self.sparse_pruning = True
        self.unconstrained_storage = False
        self.parallel_solving = False
        self.call_depth_limit = 3
        self.iprof = True
        self.solver_log = None


args = Args()



import pkg_resources
from mythril.support.support_utils import Singleton
from mythril.plugin.interface import MythrilPlugin

from typing import List, Dict, Any, Optional


class PluginDiscovery(object, metaclass=Singleton):
    """PluginDiscovery class

     This plugin implements the logic to discover and build plugins in installed python packages
    """

    # Installed plugins structure. Retrieves all modules that have an entry point for mythril.plugins
    _installed_plugins = None  # type: Optional[Dict[str, Any]]

    def init_installed_plugins(self):
        self._installed_plugins = {
            entry_point.name: entry_point.load()
            for entry_point in pkg_resources.iter_entry_points("mythril.plugins")
        }

    @property
    def installed_plugins(self):
        if self._installed_plugins is None:
            self.init_installed_plugins()
        return self._installed_plugins

    def is_installed(self, plugin_name: str) -> bool:
        """Returns whether there is python package with a plugin with plugin_name"""
        return plugin_name in self.installed_plugins.keys()

    def build_plugin(self, plugin_name: str, plugin_args: Dict) -> MythrilPlugin:
        """Returns the plugin for the given plugin_name  if it is installed"""
        if not self.is_installed(plugin_name):
            raise ValueError(f"Plugin with name: `{plugin_name}` is not installed")

        plugin = self.installed_plugins.get(plugin_name)

        if plugin is None or not issubclass(plugin, MythrilPlugin):
            raise ValueError(f"No valid plugin was found for {plugin_name}")

        return plugin(**plugin_args)

    def get_plugins(self, default_enabled=None) -> List[str]:
        """ Gets a list of installed mythril plugins

        :param default_enabled: Select plugins that are enabled by default
        :return: List of plugin names
        """
        if default_enabled is None:
            return list(self.installed_plugins.keys())

        return [
            plugin_name
            for plugin_name, plugin_class in self.installed_plugins.items()
            if plugin_class.plugin_default_enabled == default_enabled
        ]



import pytest
from mythril.disassembler.disassembly import Disassembly
from mythril.laser.ethereum.state.environment import Environment
from mythril.laser.ethereum.state.machine_state import MachineState
from mythril.laser.ethereum.state.global_state import GlobalState
from mythril.laser.ethereum.state.world_state import WorldState
from mythril.laser.ethereum.instructions import Instruction
from mythril.laser.ethereum.transaction.transaction_models import (
    MessageCallTransaction,
    TransactionStartSignal,
)
from mythril.laser.smt import symbol_factory
from mythril.laser.ethereum.state.calldata import ConcreteCalldata
from mythril.support.support_utils import get_code_hash

last_state = None
created_contract_account = None


def generate_salted_address(code_str, salt, caller):
    addr = hex(caller.value)[2:]
    addr = "0" * (40 - len(addr)) + addr

    salt = hex(salt.value)[2:]
    salt = "0" * (64 - len(salt)) + salt

    contract_address = int(
        get_code_hash("0xff" + addr + salt + get_code_hash(code_str)[2:])[26:], 16
    )
    return contract_address


def test_create2():
    world_state = WorldState()
    account = world_state.create_account(balance=10, address=101)
    account.code = Disassembly("60606040")
    environment = Environment(account, None, None, None, None, None)
    og_state = GlobalState(
        world_state, environment, None, MachineState(gas_limit=8000000)
    )
    code_raw = []
    code = "606060606060"
    for i in range(len(code) // 2):
        code_raw.append(int(code[2 * i : 2 * (i + 1)], 16))
    calldata = ConcreteCalldata("1", code_raw)
    environment.calldata = calldata
    og_state.transaction_stack.append(
        (MessageCallTransaction(world_state=WorldState(), gas_limit=8000000), None)
    )
    value = symbol_factory.BitVecVal(3, 256)
    salt = symbol_factory.BitVecVal(10, 256)
    og_state.mstate.stack = [
        salt,
        symbol_factory.BitVecVal(6, 256),
        symbol_factory.BitVecVal(0, 256),
        value,
    ]
    og_state.mstate.memory.extend(100)
    og_state.mstate.memory[0:6] = [96] * 6
    instruction = Instruction("create2", dynamic_loader=None)
    # Act + Assert
    with pytest.raises(TransactionStartSignal) as t:
        _ = instruction.evaluate(og_state)[0]
    assert t.value.transaction.call_value == value
    assert t.value.transaction.code.bytecode == code
    assert t.value.transaction.callee_account.address == generate_salted_address(
        code, salt, account.address
    )



from mythril.laser.plugin.signals import PluginSkipWorldState
from mythril.laser.plugin.interface import LaserPlugin
from mythril.laser.plugin.builder import PluginBuilder
from mythril.laser.plugin.plugins.plugin_annotations import MutationAnnotation
from mythril.laser.ethereum.state.global_state import GlobalState
from mythril.laser.ethereum.svm import LaserEVM
from mythril.laser.smt import UGT, symbol_factory
from mythril.laser.ethereum.transaction.transaction_models import (
    ContractCreationTransaction,
)
from mythril.analysis import solver
from mythril.exceptions import UnsatError


class MutationPrunerBuilder(PluginBuilder):
    name = "mutation-pruner"

    def __call__(self, *args, **kwargs):
        return MutationPruner()


class MutationPruner(LaserPlugin):
    """Mutation pruner plugin

    Let S be a world state from which T is a symbolic transaction, and S' is the resulting world state.
    In a situation where T does not execute any mutating instructions we can safely abandon further analysis on top of
    state S'. This is for the reason that we already performed analysis on S, which is equivalent.

    This optimization inhibits path explosion caused by "clean" behaviour

    The basic operation of this plugin is as follows:
     - Hook all mutating operations and introduce a MutationAnnotation to the global state on execution of the hook
     - Hook the svm EndTransaction on execution filter the states that do not have a mutation annotation

    """

    def initialize(self, symbolic_vm: LaserEVM):
        """Initializes the mutation pruner

        Introduces hooks for SSTORE operations
        :param symbolic_vm:
        :return:
        """

        @symbolic_vm.pre_hook("SSTORE")
        def sstore_mutator_hook(global_state: GlobalState):
            global_state.annotate(MutationAnnotation())

        """FIXME: Check for changes in world_state.balances instead of adding MutationAnnotation for all calls.
           Requires world_state.starting_balances to be initalized at every tx start *after* call value has been added.
        """

        @symbolic_vm.pre_hook("CALL")
        def call_mutator_hook(global_state: GlobalState):
            global_state.annotate(MutationAnnotation())

        @symbolic_vm.pre_hook("STATICCALL")
        def staticcall_mutator_hook(global_state: GlobalState):
            global_state.annotate(MutationAnnotation())

        @symbolic_vm.laser_hook("add_world_state")
        def world_state_filter_hook(global_state: GlobalState):

            if isinstance(
                global_state.current_transaction, ContractCreationTransaction
            ):
                return

            if isinstance(global_state.environment.callvalue, int):
                callvalue = symbol_factory.BitVecVal(
                    global_state.environment.callvalue, 256
                )
            else:
                callvalue = global_state.environment.callvalue

            try:

                constraints = global_state.world_state.constraints + [
                    UGT(callvalue, symbol_factory.BitVecVal(0, 256))
                ]

                solver.get_model(constraints)
                return
            except UnsatError:
                # callvalue is constrained to 0, therefore there is no balance based world state mutation
                pass

            if len(list(global_state.get_annotations(MutationAnnotation))) == 0:
                raise PluginSkipWorldState



from mythril.laser.ethereum.svm import LaserEVM
from mythril.laser.plugin.interface import LaserPlugin
from mythril.laser.plugin.builder import PluginBuilder
from time import time
import matplotlib.pyplot as plt
import logging

log = logging.getLogger(__name__)


class BenchmarkPluginBuilder(PluginBuilder):
    name = "benchmark"

    def __call__(self, *args, **kwargs):
        return BenchmarkPlugin()


# TODO: introduce dependency on coverage plugin
class BenchmarkPlugin(LaserPlugin):
    """Benchmark Plugin

    This plugin aggregates the following information:
    - duration
    - code coverage over time
    - final code coverage
    - total number of executed instructions

    """

    def __init__(self, name=None):
        """Creates BenchmarkPlugin

        :param name: name of this benchmark, used for storing the results
        """
        self.nr_of_executed_insns = 0
        self.begin = None
        self.end = None
        self.coverage = {}
        self.name = name

    def initialize(self, symbolic_vm: LaserEVM):
        """Initializes the BenchmarkPlugin

        Introduces hooks in symbolic_vm to track the desired values
        :param symbolic_vm: Symbolic virtual machine to analyze
        """
        self._reset()

        @symbolic_vm.laser_hook("execute_state")
        def execute_state_hook(_):
            current_time = time() - self.begin
            self.nr_of_executed_insns += 1

            for key, value in symbolic_vm.coverage.items():
                try:
                    self.coverage[key][current_time] = sum(value[1]) * 100 / value[0]
                except KeyError:
                    self.coverage[key] = {}
                    self.coverage[key][current_time] = sum(value[1]) * 100 / value[0]

        @symbolic_vm.laser_hook("start_sym_exec")
        def start_sym_exec_hook():
            self.begin = time()

        @symbolic_vm.laser_hook("stop_sym_exec")
        def stop_sym_exec_hook():
            self.end = time()

            self._write_to_graph()
            self._store_report()

    def _reset(self):
        """Reset this plugin"""
        self.nr_of_executed_insns = 0
        self.begin = None
        self.end = None
        self.coverage = {}

    def _store_report(self):
        """Store the results of this plugin"""
        pass

    def _write_to_graph(self):
        """Write the coverage results to a graph"""
        traces = []
        for byte_code, trace_data in self.coverage.items():
            traces += [list(trace_data.keys()), list(trace_data.values()), "r--"]

        plt.plot(*traces)
        plt.axis([0, self.end - self.begin, 0, 100])
        plt.xlabel("Duration (seconds)")
        plt.ylabel("Coverage (percentage)")

        plt.savefig("{}.png".format(self.name))



from mythril.analysis.module.base import EntryPoint, DetectionModule
from mythril.analysis.module.loader import ModuleLoader
from mythril.analysis.module.util import (
    get_detection_module_hooks,
    reset_callback_modules,
)



import pytest
from mythril.laser.ethereum.state.calldata import ConcreteCalldata, SymbolicCalldata
from mythril.laser.smt import Solver, symbol_factory
from z3 import sat, unsat
from z3.z3types import Z3Exception
from mock import MagicMock

uninitialized_test_data = [
    ([]),  # Empty concrete calldata
    ([1, 4, 5, 3, 4, 72, 230, 53]),  # Concrete calldata
]


@pytest.mark.parametrize("starting_calldata", uninitialized_test_data)
def test_concrete_calldata_uninitialized_index(starting_calldata):
    # Arrange
    calldata = ConcreteCalldata(0, starting_calldata)

    # Act
    value = calldata[100]
    value2 = calldata.get_word_at(200)

    # Assert
    assert value == 0
    assert value2 == 0


def test_concrete_calldata_calldatasize():
    # Arrange
    calldata = ConcreteCalldata(0, [1, 4, 7, 3, 7, 2, 9])
    solver = Solver()

    # Act
    solver.check()
    model = solver.model()
    result = model.eval(calldata.calldatasize.raw)

    # Assert
    assert result == 7


def test_concrete_calldata_constrain_index():
    # Arrange
    calldata = ConcreteCalldata(0, [1, 4, 7, 3, 7, 2, 9])
    solver = Solver()

    # Act
    value = calldata[2]
    constraint = value == 3

    solver.add(constraint)
    result = solver.check()

    # Assert
    assert str(result) == "unsat"


def test_symbolic_calldata_constrain_index():
    # Arrange
    calldata = SymbolicCalldata(0)
    solver = Solver()

    # Act
    value = calldata[51]

    constraints = [value == 1, calldata.calldatasize == 50]

    solver.add(*constraints)

    result = solver.check()

    # Assert
    assert str(result) == "unsat"


def test_symbolic_calldata_equal_indices():
    calldata = SymbolicCalldata(0)

    index_a = symbol_factory.BitVecSym("index_a", 256)
    index_b = symbol_factory.BitVecSym("index_b", 256)

    # Act
    a = calldata[index_a]
    b = calldata[index_b]

    s = Solver()
    s.append(index_a == index_b)
    s.append(a != b)

    # Assert
    assert unsat == s.check()



from time import time
from mythril.support.support_utils import Singleton


class StartTime(metaclass=Singleton):
    """Maintains the start time of the current contract in execution"""

    def __init__(self):
        self.global_start_time = time()



"""This module contains various utility conversion functions and constants for
LASER."""
import re
from typing import Dict, List, Union, TYPE_CHECKING, cast

if TYPE_CHECKING:
    from mythril.laser.ethereum.state.machine_state import MachineState

from mythril.laser.smt import BitVec, Bool, Expression, If, simplify, symbol_factory

TT256 = 2 ** 256
TT256M1 = 2 ** 256 - 1
TT255 = 2 ** 255


def safe_decode(hex_encoded_string: str) -> bytes:
    """

    :param hex_encoded_string:
    :return:
    """
    if hex_encoded_string.startswith("0x"):
        return bytes.fromhex(hex_encoded_string[2:])
    else:
        return bytes.fromhex(hex_encoded_string)


def to_signed(i: int) -> int:
    """

    :param i:
    :return:
    """
    return i if i < TT255 else i - TT256


def get_instruction_index(
    instruction_list: List[Dict], address: int
) -> Union[int, None]:
    """

    :param instruction_list:
    :param address:
    :return:
    """
    index = 0
    for instr in instruction_list:
        if instr["address"] >= address:
            return index
        index += 1
    return None


def get_trace_line(instr: Dict, state: "MachineState") -> str:
    """

    :param instr:
    :param state:
    :return:
    """
    stack = str(state.stack[::-1])
    # stack = re.sub("(\d+)",   lambda m: hex(int(m.group(1))), stack)
    stack = re.sub("\n", "", stack)
    return str(instr["address"]) + " " + instr["opcode"] + "\tSTACK: " + stack


def pop_bitvec(state: "MachineState") -> BitVec:
    """

    :param state:
    :return:
    """
    # pop one element from stack, converting boolean expressions and
    # concrete Python variables to BitVecVal

    item = state.stack.pop()

    if isinstance(item, Bool):
        return If(
            cast(Bool, item),
            symbol_factory.BitVecVal(1, 256),
            symbol_factory.BitVecVal(0, 256),
        )
    elif isinstance(item, int):
        return symbol_factory.BitVecVal(item, 256)
    else:
        item = cast(BitVec, item)
        return simplify(item)


def get_concrete_int(item: Union[int, Expression]) -> int:
    """

    :param item:
    :return:
    """
    if isinstance(item, int):
        return item
    elif isinstance(item, BitVec):
        if item.symbolic:
            raise TypeError("Got a symbolic BitVecRef")
        return item.value
    elif isinstance(item, Bool):
        value = item.value
        if value is None:
            raise TypeError("Symbolic boolref encountered")
        return value

    assert False, "Unhandled type {} encountered".format(str(type(item)))


def concrete_int_from_bytes(
    concrete_bytes: Union[List[Union[BitVec, int]], bytes], start_index: int
) -> int:
    """

    :param concrete_bytes:
    :param start_index:
    :return:
    """
    concrete_bytes = [
        byte.value if isinstance(byte, BitVec) and not byte.symbolic else byte
        for byte in concrete_bytes
    ]
    integer_bytes = concrete_bytes[start_index : start_index + 32]

    # The below statement is expected to fail in some circumstances whose error is caught
    return int.from_bytes(integer_bytes, byteorder="big")  # type: ignore


def concrete_int_to_bytes(val):
    """

    :param val:
    :return:
    """
    # logging.debug("concrete_int_to_bytes " + str(val))
    if type(val) == int:
        return val.to_bytes(32, byteorder="big")
    return simplify(val).value.to_bytes(32, byteorder="big")


def bytearray_to_int(arr):
    """

    :param arr:
    :return:
    """
    o = 0
    for a in arr:
        o = (o << 8) + a
    return o


def extract_copy(
    data: bytearray, mem: bytearray, memstart: int, datastart: int, size: int
):
    for i in range(size):
        if datastart + i < len(data):
            mem[memstart + i] = data[datastart + i]
        else:
            mem[memstart + i] = 0


def extract32(data: bytearray, i: int) -> int:
    """

    :param data:
    :param i:
    :return:
    """
    if i >= len(data):
        return 0
    o = data[i : min(i + 32, len(data))]
    o.extend(bytearray(32 - len(o)))
    return bytearray_to_int(o)



# -*- coding: utf-8 -*-
"""install mythril and deploy source-dist and wheel to pypi.python.org.

deps (requires up2date version):
    *) pip install --upgrade pip wheel setuptools twine
publish to pypi w/o having to convert Readme.md to RST:
    1) #> python setup.py sdist bdist_wheel
    2) #> twine upload dist/*   #<specify bdist_wheel version to upload>; #optional --repository <testpypi> or  --repository-url <testpypi-url>
"""
from setuptools import setup, find_packages
from setuptools.command.install import install
import sys
import os
import io

# Package meta-data.
NAME = "mythril"
DESCRIPTION = "Security analysis tool for Ethereum smart contracts"
URL = "https://github.com/ConsenSys/mythril"
AUTHOR = "ConsenSys Dilligence"
AUTHOR_MAIL = None
REQUIRES_PYTHON = ">=3.6.0"


# What packages are required for this module to be executed?
REQUIRED = [
    "coloredlogs>=10.0",
    "py_ecc==1.6.0",
    "ethereum>=2.3.2",
    "z3-solver>=4.8.5.0",
    "requests>=2.22.0",
    "py-solc",
    "py-solc-x==1.0.0",
    "semantic_version==2.8.5",
    "plyvel",
    "eth_abi==1.3.0",
    "eth-account>=0.1.0a2,<=0.3.0",
    "eth-hash>=0.1.0",
    "eth-keyfile>=0.5.1",
    "eth-keys>=0.2.0b3,<0.3.0",
    "eth-rlp>=0.1.0",
    "eth-tester==0.1.0b32",
    "eth-typing>=2.0.0",
    "coverage",
    "jinja2>=2.9",
    "rlp>=1.0.1,<2.0.0",
    "transaction>=2.2.1",
    "py-flags",
    "mock",
    "configparser>=3.5.0",
    "persistent>=4.2.0",
    "ethereum-input-decoder>=0.2.2",
    "matplotlib",
    "pythx",
    "py-evm==0.3.0a13",
    "certifi>=2020.06.20",
]

TESTS_REQUIRE = ["mypy", "pytest>=3.6.0", "pytest_mock", "pytest-cov"]

# What packages are optional?
EXTRAS = {
    # 'fancy feature': ['django'],
}

# If version is set to None then it will be fetched from __version__.py
VERSION = None

here = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
# Note: this will only work if 'README.md' is present in your MANIFEST.in file!
try:
    with io.open(os.path.join(here, "README.md"), encoding="utf-8") as f:
        long_description = "\n" + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION


# Load the package's __version__.py module as a dictionary.
about = {}
if not VERSION:
    project_slug = NAME.lower().replace("-", "_").replace(" ", "_")
    with open(os.path.join(here, project_slug, "__version__.py")) as f:
        exec(f.read(), about)
else:
    about["__version__"] = VERSION


# Package version (vX.Y.Z). It must match git tag being used for CircleCI
# deployment; otherwise the build will failed.
class VerifyVersionCommand(install):
    """Custom command to verify that the git tag matches our version."""

    description = "verify that the git tag matches our version"

    def run(self):
        """"""
        tag = os.getenv("CIRCLE_TAG")

        if tag != about["__version__"]:
            info = "Git tag: {0} does not match the version of this app: {1}".format(
                tag, about["__version__"]
            )
            sys.exit(info)


setup(
    name=NAME,
    version=about["__version__"][1:],
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/markdown",  # requires twine and recent setuptools
    url=URL,
    author=AUTHOR,
    author_mail=AUTHOR_MAIL,
    license="MIT",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "Topic :: Software Development :: Disassemblers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    keywords="hacking disassembler security ethereum",
    packages=find_packages(exclude=["contrib", "docs", "tests"]),
    install_requires=REQUIRED,
    tests_require=TESTS_REQUIRE,
    python_requires=REQUIRES_PYTHON,
    extras_require=EXTRAS,
    package_data={"mythril.analysis.templates": ["*"], "mythril.support.assets": ["*"]},
    include_package_data=True,
    entry_points={"console_scripts": ["myth=mythril.interfaces.cli:main"]},
    cmdclass={"verify": VerifyVersionCommand},
)



"""This module contains various helper methods for dealing with EVM
operations."""
from enum import Enum

from mythril.laser.ethereum import util
from mythril.laser.smt import simplify


class VarType(Enum):
    """An enum denoting whether a value is symbolic or concrete."""

    SYMBOLIC = 1
    CONCRETE = 2


class Variable:
    """The representation of a variable with value and type."""

    def __init__(self, val, _type):
        """

        :param val:
        :param _type:
        """
        self.val = val
        self.type = _type

    def __str__(self):
        """

        :return:
        """
        return str(self.val)


def get_variable(i):
    """

    :param i:
    :return:
    """
    try:
        return Variable(util.get_concrete_int(i), VarType.CONCRETE)
    except TypeError:
        return Variable(simplify(i), VarType.SYMBOLIC)


class Op:
    """The base type for operations referencing current node and state."""

    def __init__(self, node, state, state_index):
        """

        :param node:
        :param state:
        :param state_index:
        """
        self.node = node
        self.state = state
        self.state_index = state_index


class Call(Op):
    """The representation of a CALL operation."""

    def __init__(
        self,
        node,
        state,
        state_index,
        _type,
        to,
        gas,
        value=Variable(0, VarType.CONCRETE),
        data=None,
    ):
        """

        :param node:
        :param state:
        :param state_index:
        :param _type:
        :param to:
        :param gas:
        :param value:
        :param data:
        """
        super().__init__(node, state, state_index)
        self.to = to
        self.gas = gas
        self.type = _type
        self.value = value
        self.data = data



"""This module contains analysis module helpers to solve path constraints."""
from typing import Dict, List, Tuple, Union
from z3 import FuncInterp
import z3

from mythril.support.model import get_model
from mythril.laser.ethereum.state.global_state import GlobalState
from mythril.laser.ethereum.state.constraints import Constraints
from mythril.laser.ethereum.keccak_function_manager import (
    keccak_function_manager,
    hash_matcher,
)
from mythril.laser.ethereum.transaction import BaseTransaction
from mythril.laser.smt import UGE, symbol_factory
from mythril.exceptions import UnsatError
from mythril.laser.ethereum.transaction.transaction_models import (
    ContractCreationTransaction,
)
import logging

log = logging.getLogger(__name__)


def pretty_print_model(model):
    """ Pretty prints a z3 model

    :param model:
    :return:
    """
    ret = ""

    for d in model.decls():
        if type(model[d]) == FuncInterp:
            condition = model[d].as_list()
            ret += "%s: %s\n" % (d.name(), condition)
            continue

        try:
            condition = "0x%x" % model[d].as_long()
        except:
            condition = str(z3.simplify(model[d]))

        ret += "%s: %s\n" % (d.name(), condition)

    return ret


def get_transaction_sequence(
    global_state: GlobalState, constraints: Constraints
) -> Dict:
    """Generate concrete transaction sequence.

    :param global_state: GlobalState to generate transaction sequence for
    :param constraints: list of constraints used to generate transaction sequence
    """

    transaction_sequence = global_state.world_state.transaction_sequence

    concrete_transactions = []

    tx_constraints, minimize = _set_minimisation_constraints(
        transaction_sequence, constraints.copy(), [], 5000, global_state.world_state
    )
    try:
        model = get_model(tx_constraints, minimize=minimize)
    except UnsatError:
        raise UnsatError

    # Include creation account in initial state
    # Note: This contains the code, which should not exist until after the first tx
    initial_world_state = transaction_sequence[0].world_state
    initial_accounts = initial_world_state.accounts

    for transaction in transaction_sequence:
        concrete_transaction = _get_concrete_transaction(model, transaction)
        concrete_transactions.append(concrete_transaction)

    min_price_dict = {}  # type: Dict[str, int]
    for address in initial_accounts.keys():
        min_price_dict[address] = model.eval(
            initial_world_state.starting_balances[
                symbol_factory.BitVecVal(address, 256)
            ].raw,
            model_completion=True,
        ).as_long()

    concrete_initial_state = _get_concrete_state(initial_accounts, min_price_dict)
    if isinstance(transaction_sequence[0], ContractCreationTransaction):
        code = transaction_sequence[0].code
        _replace_with_actual_sha(concrete_transactions, model, code)
    else:
        _replace_with_actual_sha(concrete_transactions, model)
    _add_calldata_placeholder(concrete_transactions, transaction_sequence)
    steps = {"initialState": concrete_initial_state, "steps": concrete_transactions}

    return steps


def _add_calldata_placeholder(
    concrete_transactions: List[Dict[str, str]],
    transaction_sequence: List[BaseTransaction],
):
    """
    Add's a calldata placeholder into the concrete transactions
    :param concrete_transactions:
    :param transaction_sequence:
    :return:
    """
    for tx in concrete_transactions:
        tx["calldata"] = tx["input"]
    if not isinstance(transaction_sequence[0], ContractCreationTransaction):
        return
    code_len = len(transaction_sequence[0].code.bytecode)
    concrete_transactions[0]["calldata"] = concrete_transactions[0]["input"][
        code_len + 2 :
    ]


def _replace_with_actual_sha(
    concrete_transactions: List[Dict[str, str]], model: z3.Model, code=None
):
    concrete_hashes = keccak_function_manager.get_concrete_hash_data(model)
    for tx in concrete_transactions:
        if hash_matcher not in tx["input"]:
            continue
        if code is not None and code.bytecode in tx["input"]:
            s_index = len(code.bytecode) + 2
        else:
            s_index = 10
        for i in range(s_index, len(tx["input"])):
            data_slice = tx["input"][i : i + 64]
            if hash_matcher not in data_slice or len(data_slice) != 64:
                continue
            find_input = symbol_factory.BitVecVal(int(data_slice, 16), 256)
            input_ = None
            for size in concrete_hashes:
                _, inverse = keccak_function_manager.store_function[size]
                if find_input.value not in concrete_hashes[size]:
                    continue
                input_ = symbol_factory.BitVecVal(
                    model.eval(inverse(find_input).raw).as_long(), size
                )

            if input_ is None:
                continue
            keccak = keccak_function_manager.find_concrete_keccak(input_)
            hex_keccak = hex(keccak.value)[2:]
            if len(hex_keccak) != 64:
                hex_keccak = "0" * (64 - len(hex_keccak)) + hex_keccak
            tx["input"] = tx["input"][:s_index] + tx["input"][s_index:].replace(
                tx["input"][i : 64 + i], hex_keccak
            )


def _get_concrete_state(initial_accounts: Dict, min_price_dict: Dict[str, int]):
    """ Gets a concrete state """
    accounts = {}
    for address, account in initial_accounts.items():
        # Skip empty default account

        data = dict()  # type: Dict[str, Union[int, str]]
        data["nonce"] = account.nonce
        data["code"] = account.code.bytecode
        data["storage"] = str(account.storage)
        data["balance"] = hex(min_price_dict.get(address, 0))
        accounts[hex(address)] = data
    return {"accounts": accounts}


def _get_concrete_transaction(model: z3.Model, transaction: BaseTransaction):
    """ Gets a concrete transaction from a transaction and z3 model"""
    # Get concrete values from transaction
    address = hex(transaction.callee_account.address.value)
    value = model.eval(transaction.call_value.raw, model_completion=True).as_long()
    caller = "0x" + (
        "%x" % model.eval(transaction.caller.raw, model_completion=True).as_long()
    ).zfill(40)

    input_ = ""
    if isinstance(transaction, ContractCreationTransaction):
        address = ""
        input_ += transaction.code.bytecode

    input_ += "".join(
        [
            hex(b)[2:] if len(hex(b)) % 2 == 0 else "0" + hex(b)[2:]
            for b in transaction.call_data.concrete(model)
        ]
    )

    # Create concrete transaction dict
    concrete_transaction = dict()  # type: Dict[str, str]
    concrete_transaction["input"] = "0x" + input_
    concrete_transaction["value"] = "0x%x" % value
    # Fixme: base origin assignment on origin symbol
    concrete_transaction["origin"] = caller
    concrete_transaction["address"] = "%s" % address

    return concrete_transaction


def _set_minimisation_constraints(
    transaction_sequence, constraints, minimize, max_size, world_state
) -> Tuple[Constraints, tuple]:
    """ Set constraints that minimise key transaction values

    Constraints generated:
    - Upper bound on calldata size
    - Minimisation of call value's and calldata sizes

    :param transaction_sequence: Transaction for which the constraints should be applied
    :param constraints: The constraints array which should contain any added constraints
    :param minimize: The minimisation array which should contain any variables that should be minimised
    :param max_size: The max size of the calldata array
    :return: updated constraints, minimize
    """
    for transaction in transaction_sequence:
        # Set upper bound on calldata size
        max_calldata_size = symbol_factory.BitVecVal(max_size, 256)
        constraints.append(UGE(max_calldata_size, transaction.call_data.calldatasize))

        # Minimize
        minimize.append(transaction.call_data.calldatasize)
        minimize.append(transaction.call_value)
        constraints.append(
            UGE(
                symbol_factory.BitVecVal(1000000000000000000000, 256),
                world_state.starting_balances[transaction.caller],
            )
        )

    for account in world_state.accounts.values():
        # Lazy way to prevent overflows and to ensure "reasonable" balances
        # Each account starts with less than 100 ETH
        constraints.append(
            UGE(
                symbol_factory.BitVecVal(100000000000000000000, 256),
                world_state.starting_balances[account.address],
            )
        )

    return constraints, tuple(minimize)



from ethereum import opcodes
from ethereum.utils import ceil32
from typing import Callable, Dict, Tuple, Union


Z_OPERATOR_TUPLE = (0, 1)
UN_OPERATOR_TUPLE = (1, 1)
BIN_OPERATOR_TUPLE = (2, 1)
T_OPERATOR_TUPLE = (3, 1)
GAS = "gas"
STACK = "stack"

# Gas tuple contains (min_gas, max_gas)
# stack tuple contains (no_of_elements_popped, no_of_elements_pushed)

OPCODES = {
    "STOP": {GAS: (0, 0), STACK: (0, 0)},
    "ADD": {GAS: (3, 3), STACK: BIN_OPERATOR_TUPLE},
    "MUL": {GAS: (5, 5), STACK: BIN_OPERATOR_TUPLE},
    "SUB": {GAS: (3, 3), STACK: BIN_OPERATOR_TUPLE},
    "DIV": {GAS: (5, 5), STACK: BIN_OPERATOR_TUPLE},
    "SDIV": {GAS: (5, 5), STACK: BIN_OPERATOR_TUPLE},
    "MOD": {GAS: (5, 5), STACK: BIN_OPERATOR_TUPLE},
    "SMOD": {GAS: (5, 5), STACK: BIN_OPERATOR_TUPLE},
    "ADDMOD": {GAS: (8, 8), STACK: BIN_OPERATOR_TUPLE},
    "MULMOD": {GAS: (8, 8), STACK: T_OPERATOR_TUPLE},
    "EXP": {GAS: (10, 340), STACK: BIN_OPERATOR_TUPLE},  # exponent max 2^32
    "SIGNEXTEND": {GAS: (5, 5), STACK: BIN_OPERATOR_TUPLE},
    "LT": {GAS: (3, 3), STACK: BIN_OPERATOR_TUPLE},
    "GT": {GAS: (3, 3), STACK: BIN_OPERATOR_TUPLE},
    "SLT": {GAS: (3, 3), STACK: BIN_OPERATOR_TUPLE},
    "SGT": {GAS: (3, 3), STACK: BIN_OPERATOR_TUPLE},
    "EQ": {GAS: (3, 3), STACK: BIN_OPERATOR_TUPLE},
    "AND": {GAS: (3, 3), STACK: BIN_OPERATOR_TUPLE},
    "ISZERO": {GAS: (3, 3), STACK: UN_OPERATOR_TUPLE},
    "OR": {GAS: (3, 3), STACK: BIN_OPERATOR_TUPLE},
    "XOR": {GAS: (3, 3), STACK: BIN_OPERATOR_TUPLE},
    "NOT": {GAS: (3, 3), STACK: UN_OPERATOR_TUPLE},
    "BYTE": {GAS: (3, 3), STACK: BIN_OPERATOR_TUPLE},
    "SHL": {GAS: (3, 3), STACK: BIN_OPERATOR_TUPLE},
    "SHR": {GAS: (3, 3), STACK: BIN_OPERATOR_TUPLE},
    "SAR": {GAS: (3, 3), STACK: BIN_OPERATOR_TUPLE},
    "SHA3": {
        GAS: (
            30,
            30 + 6 * 8,
        ),  # max can be larger, but usually storage location with 8 words
        STACK: BIN_OPERATOR_TUPLE,
    },
    "ADDRESS": {GAS: (2, 2), STACK: Z_OPERATOR_TUPLE},
    "BALANCE": {GAS: (700, 700), STACK: UN_OPERATOR_TUPLE},
    "ORIGIN": {GAS: (2, 2), STACK: Z_OPERATOR_TUPLE},
    "CALLER": {GAS: (2, 2), STACK: Z_OPERATOR_TUPLE},
    "CALLVALUE": {GAS: (2, 2), STACK: Z_OPERATOR_TUPLE},
    "CALLDATALOAD": {GAS: (3, 3), STACK: UN_OPERATOR_TUPLE},
    "CALLDATASIZE": {GAS: (2, 2), STACK: Z_OPERATOR_TUPLE},
    "CALLDATACOPY": {
        GAS: (2, 2 + 3 * 768),  # https://ethereum.stackexchange.com/a/47556
        STACK: (3, 0),
    },
    "CODESIZE": {GAS: (2, 2), STACK: Z_OPERATOR_TUPLE},
    "CODECOPY": {
        GAS: (2, 2 + 3 * 768),  # https://ethereum.stackexchange.com/a/47556,
        STACK: (3, 0),
    },
    "GASPRICE": {GAS: (2, 2), STACK: Z_OPERATOR_TUPLE},
    "EXTCODESIZE": {GAS: (700, 700), STACK: Z_OPERATOR_TUPLE},
    "EXTCODECOPY": {
        GAS: (700, 700 + 3 * 768),  # https://ethereum.stackexchange.com/a/47556
        STACK: (4, 0),
    },
    "EXTCODEHASH": {GAS: (700, 700), STACK: UN_OPERATOR_TUPLE},
    "RETURNDATASIZE": {GAS: (2, 2), STACK: Z_OPERATOR_TUPLE},
    "RETURNDATACOPY": {GAS: (3, 3), STACK: (3, 0)},
    "BLOCKHASH": {GAS: (20, 20), STACK: UN_OPERATOR_TUPLE},
    "COINBASE": {GAS: (2, 2), STACK: Z_OPERATOR_TUPLE},
    "TIMESTAMP": {GAS: (2, 2), STACK: Z_OPERATOR_TUPLE},
    "NUMBER": {GAS: (2, 2), STACK: Z_OPERATOR_TUPLE},
    "DIFFICULTY": {GAS: (2, 2), STACK: Z_OPERATOR_TUPLE},
    "GASLIMIT": {GAS: (2, 2), STACK: Z_OPERATOR_TUPLE},
    "CHAINID": {GAS: (2, 2), STACK: Z_OPERATOR_TUPLE},
    "SELFBALANCE": {GAS: (2, 2), STACK: Z_OPERATOR_TUPLE},
    "POP": {GAS: (2, 2), STACK: (1, 0)},
    # assume 1KB memory r/w cost as upper bound
    "MLOAD": {GAS: (3, 96), STACK: UN_OPERATOR_TUPLE},
    "MSTORE": {GAS: (3, 98), STACK: (2, 0)},
    "MSTORE8": {GAS: (3, 98), STACK: (2, 0)},
    # assume 64 byte r/w cost as upper bound
    "SLOAD": {GAS: (800, 800), STACK: UN_OPERATOR_TUPLE},
    "SSTORE": {GAS: (5000, 25000), STACK: (1, 0)},
    "JUMP": {GAS: (8, 8), STACK: (1, 0)},
    "JUMPI": {GAS: (10, 10), STACK: (2, 0)},
    "PC": {GAS: (2, 2), STACK: Z_OPERATOR_TUPLE},
    "MSIZE": {GAS: (2, 2), STACK: Z_OPERATOR_TUPLE},
    "GAS": {GAS: (2, 2), STACK: Z_OPERATOR_TUPLE},
    "JUMPDEST": {GAS: (1, 1), STACK: (0, 0)},
    "PUSH1": {GAS: (3, 3), STACK: Z_OPERATOR_TUPLE},
    "PUSH2": {GAS: (3, 3), STACK: Z_OPERATOR_TUPLE},
    "PUSH3": {GAS: (3, 3), STACK: Z_OPERATOR_TUPLE},
    "PUSH4": {GAS: (3, 3), STACK: Z_OPERATOR_TUPLE},
    "PUSH5": {GAS: (3, 3), STACK: Z_OPERATOR_TUPLE},
    "PUSH6": {GAS: (3, 3), STACK: Z_OPERATOR_TUPLE},
    "PUSH7": {GAS: (3, 3), STACK: Z_OPERATOR_TUPLE},
    "PUSH8": {GAS: (3, 3), STACK: Z_OPERATOR_TUPLE},
    "PUSH9": {GAS: (3, 3), STACK: Z_OPERATOR_TUPLE},
    "PUSH10": {GAS: (3, 3), STACK: Z_OPERATOR_TUPLE},
    "PUSH11": {GAS: (3, 3), STACK: Z_OPERATOR_TUPLE},
    "PUSH12": {GAS: (3, 3), STACK: Z_OPERATOR_TUPLE},
    "PUSH13": {GAS: (3, 3), STACK: Z_OPERATOR_TUPLE},
    "PUSH14": {GAS: (3, 3), STACK: Z_OPERATOR_TUPLE},
    "PUSH15": {GAS: (3, 3), STACK: Z_OPERATOR_TUPLE},
    "PUSH16": {GAS: (3, 3), STACK: Z_OPERATOR_TUPLE},
    "PUSH17": {GAS: (3, 3), STACK: Z_OPERATOR_TUPLE},
    "PUSH18": {GAS: (3, 3), STACK: Z_OPERATOR_TUPLE},
    "PUSH19": {GAS: (3, 3), STACK: Z_OPERATOR_TUPLE},
    "PUSH20": {GAS: (3, 3), STACK: Z_OPERATOR_TUPLE},
    "PUSH21": {GAS: (3, 3), STACK: Z_OPERATOR_TUPLE},
    "PUSH22": {GAS: (3, 3), STACK: Z_OPERATOR_TUPLE},
    "PUSH23": {GAS: (3, 3), STACK: Z_OPERATOR_TUPLE},
    "PUSH24": {GAS: (3, 3), STACK: Z_OPERATOR_TUPLE},
    "PUSH25": {GAS: (3, 3), STACK: Z_OPERATOR_TUPLE},
    "PUSH26": {GAS: (3, 3), STACK: Z_OPERATOR_TUPLE},
    "PUSH27": {GAS: (3, 3), STACK: Z_OPERATOR_TUPLE},
    "PUSH28": {GAS: (3, 3), STACK: Z_OPERATOR_TUPLE},
    "PUSH29": {GAS: (3, 3), STACK: Z_OPERATOR_TUPLE},
    "PUSH30": {GAS: (3, 3), STACK: Z_OPERATOR_TUPLE},
    "PUSH31": {GAS: (3, 3), STACK: Z_OPERATOR_TUPLE},
    "PUSH32": {GAS: (3, 3), STACK: Z_OPERATOR_TUPLE},
    "DUP1": {GAS: (3, 3), STACK: Z_OPERATOR_TUPLE},
    "DUP2": {GAS: (3, 3), STACK: Z_OPERATOR_TUPLE},
    "DUP3": {GAS: (3, 3), STACK: Z_OPERATOR_TUPLE},
    "DUP4": {GAS: (3, 3), STACK: Z_OPERATOR_TUPLE},
    "DUP5": {GAS: (3, 3), STACK: Z_OPERATOR_TUPLE},
    "DUP6": {GAS: (3, 3), STACK: Z_OPERATOR_TUPLE},
    "DUP7": {GAS: (3, 3), STACK: Z_OPERATOR_TUPLE},
    "DUP8": {GAS: (3, 3), STACK: Z_OPERATOR_TUPLE},
    "DUP9": {GAS: (3, 3), STACK: Z_OPERATOR_TUPLE},
    "DUP10": {GAS: (3, 3), STACK: Z_OPERATOR_TUPLE},
    "DUP11": {GAS: (3, 3), STACK: Z_OPERATOR_TUPLE},
    "DUP12": {GAS: (3, 3), STACK: Z_OPERATOR_TUPLE},
    "DUP13": {GAS: (3, 3), STACK: Z_OPERATOR_TUPLE},
    "DUP14": {GAS: (3, 3), STACK: Z_OPERATOR_TUPLE},
    "DUP15": {GAS: (3, 3), STACK: Z_OPERATOR_TUPLE},
    "DUP16": {GAS: (3, 3), STACK: Z_OPERATOR_TUPLE},
    "SWAP1": {GAS: (3, 3), STACK: (0, 0)},
    "SWAP2": {GAS: (3, 3), STACK: (0, 0)},
    "SWAP3": {GAS: (3, 3), STACK: (0, 0)},
    "SWAP4": {GAS: (3, 3), STACK: (0, 0)},
    "SWAP5": {GAS: (3, 3), STACK: (0, 0)},
    "SWAP6": {GAS: (3, 3), STACK: (0, 0)},
    "SWAP7": {GAS: (3, 3), STACK: (0, 0)},
    "SWAP8": {GAS: (3, 3), STACK: (0, 0)},
    "SWAP9": {GAS: (3, 3), STACK: (0, 0)},
    "SWAP10": {GAS: (3, 3), STACK: (0, 0)},
    "SWAP11": {GAS: (3, 3), STACK: (0, 0)},
    "SWAP12": {GAS: (3, 3), STACK: (0, 0)},
    "SWAP13": {GAS: (3, 3), STACK: (0, 0)},
    "SWAP14": {GAS: (3, 3), STACK: (0, 0)},
    "SWAP15": {GAS: (3, 3), STACK: (0, 0)},
    "SWAP16": {GAS: (3, 3), STACK: (0, 0)},
    # apparently Solidity only allows byte32 as input to the log
    # function. Virtually it could be as large as the block gas limit
    # allows, but let's stick to the reasonable standard here.
    # https://ethereum.stackexchange.com/a/1691
    "LOG0": {GAS: (375, 375 + 8 * 32), STACK: (2, 0)},
    "LOG1": {GAS: (2 * 375, 2 * 375 + 8 * 32), STACK: (3, 0)},
    "LOG2": {GAS: (3 * 375, 3 * 375 + 8 * 32), STACK: (4, 0)},
    "LOG3": {GAS: (4 * 375, 4 * 375 + 8 * 32), STACK: (5, 0)},
    "LOG4": {GAS: (5 * 375, 5 * 375 + 8 * 32), STACK: (6, 0)},
    "CREATE": {GAS: (32000, 32000), STACK: T_OPERATOR_TUPLE},
    "CREATE2": {
        GAS: (32000, 32000),  # TODO: Make the gas values dynamic
        STACK: (4, 1),
    },
    "CALL": {GAS: (700, 700 + 9000 + 25000), STACK: (7, 1)},
    "CALLCODE": {GAS: (700, 700 + 9000 + 25000), STACK: (7, 1)},
    "RETURN": {GAS: (0, 0), STACK: (2, 0)},
    "DELEGATECALL": {GAS: (700, 700 + 9000 + 25000), STACK: (6, 1)},
    "STATICCALL": {GAS: (700, 700 + 9000 + 25000), STACK: (6, 1)},
    "REVERT": {GAS: (0, 0), STACK: (2, 0)},
    "SUICIDE": {GAS: (5000, 30000), STACK: (1, 0)},
    "ASSERT_FAIL": {GAS: (0, 0), STACK: (0, 0)},
    "INVALID": {GAS: (0, 0), STACK: (0, 0)},
    "BEGINSUB": {GAS: (2, 2), STACK: (0, 0)},
    "JUMPSUB": {GAS: (10, 10), STACK: (1, 0)},
    "RETURNSUB": {GAS: (5, 5), STACK: (0, 0)},
}  # type: Dict[str, Dict[str, Tuple[int, int]]]


def calculate_sha3_gas(length: int):
    """

    :param length:
    :return:
    """
    gas_val = 30 + opcodes.GSHA3WORD * (ceil32(length) // 32)
    return gas_val, gas_val


def calculate_native_gas(size: int, contract: str):
    """

    :param size:
    :param contract:
    :return:
    """
    gas_value = 0
    word_num = ceil32(size) // 32
    if contract == "ecrecover":
        gas_value = opcodes.GECRECOVER
    elif contract == "sha256":
        gas_value = opcodes.GSHA256BASE + word_num * opcodes.GSHA256WORD
    elif contract == "ripemd160":
        gas_value = opcodes.GRIPEMD160BASE + word_num * opcodes.GRIPEMD160WORD
    elif contract == "identity":
        gas_value = opcodes.GIDENTITYBASE + word_num * opcodes.GIDENTITYWORD
    else:
        # TODO: Add gas for other precompiles, computation should be shifted to natives.py
        #  as some data isn't available here
        pass
    return gas_value, gas_value


def get_opcode_gas(opcode: str) -> Tuple[int, int]:
    return OPCODES[opcode][GAS]


def get_required_stack_elements(opcode: str) -> int:
    return OPCODES[opcode][STACK][0]



"""This module contains representation classes for Solidity files, contracts
and source mappings."""
from typing import Dict, Set

import mythril.laser.ethereum.util as helper
from mythril.ethereum.evmcontract import EVMContract
from mythril.ethereum.util import get_solc_json
from mythril.exceptions import NoContractFoundError


class SourceMapping:
    def __init__(self, solidity_file_idx, offset, length, lineno, mapping):
        """Representation of a source mapping for a Solidity file."""

        self.solidity_file_idx = solidity_file_idx
        self.offset = offset
        self.length = length
        self.lineno = lineno
        self.solc_mapping = mapping


class SolidityFile:
    """Representation of a file containing Solidity code."""

    def __init__(self, filename: str, data: str, full_contract_src_maps: Set[str]):
        """
        Metadata class containing data regarding a specific solidity file
        :param filename: The filename of the solidity file
        :param data: The code of the solidity file
        :param full_contract_src_maps: The set of contract source mappings of all the contracts in the file
        """
        self.filename = filename
        self.data = data
        self.full_contract_src_maps = full_contract_src_maps


class SourceCodeInfo:
    def __init__(self, filename, lineno, code, mapping):
        """Metadata class containing a code reference for a specific file."""

        self.filename = filename
        self.lineno = lineno
        self.code = code
        self.solc_mapping = mapping


def get_contracts_from_file(input_file, solc_settings_json=None, solc_binary="solc"):
    """

    :param input_file:
    :param solc_settings_json:
    :param solc_binary:
    """
    data = get_solc_json(
        input_file, solc_settings_json=solc_settings_json, solc_binary=solc_binary
    )

    try:
        for contract_name in data["contracts"][input_file].keys():
            if len(
                data["contracts"][input_file][contract_name]["evm"]["deployedBytecode"][
                    "object"
                ]
            ):
                yield SolidityContract(
                    input_file=input_file,
                    name=contract_name,
                    solc_settings_json=solc_settings_json,
                    solc_binary=solc_binary,
                )
    except KeyError:
        raise NoContractFoundError


class SolidityContract(EVMContract):
    """Representation of a Solidity contract."""

    def __init__(
        self, input_file, name=None, solc_settings_json=None, solc_binary="solc"
    ):
        data = get_solc_json(
            input_file, solc_settings_json=solc_settings_json, solc_binary=solc_binary
        )

        self.solidity_files = []
        self.solc_json = data
        self.input_file = input_file

        for filename, contract in data["sources"].items():
            with open(filename, "r", encoding="utf-8") as file:
                code = file.read()
                full_contract_src_maps = self.get_full_contract_src_maps(
                    contract["ast"]
                )
                self.solidity_files.append(
                    SolidityFile(filename, code, full_contract_src_maps)
                )

        has_contract = False

        # If a contract name has been specified, find the bytecode of that specific contract
        srcmap_constructor = []
        srcmap = []
        if name:
            contract = data["contracts"][input_file][name]
            if len(contract["evm"]["deployedBytecode"]["object"]):
                code = contract["evm"]["deployedBytecode"]["object"]
                creation_code = contract["evm"]["bytecode"]["object"]
                srcmap = contract["evm"]["deployedBytecode"]["sourceMap"].split(";")
                srcmap_constructor = contract["evm"]["bytecode"]["sourceMap"].split(";")
                has_contract = True

        # If no contract name is specified, get the last bytecode entry for the input file

        else:
            for contract_name, contract in sorted(
                data["contracts"][input_file].items()
            ):
                if len(contract["evm"]["deployedBytecode"]["object"]):
                    name = contract_name
                    code = contract["evm"]["deployedBytecode"]["object"]
                    creation_code = contract["evm"]["bytecode"]["object"]
                    srcmap = contract["evm"]["deployedBytecode"]["sourceMap"].split(";")
                    srcmap_constructor = contract["evm"]["bytecode"]["sourceMap"].split(
                        ";"
                    )
                    has_contract = True

        if not has_contract:
            raise NoContractFoundError

        self.mappings = []

        self.constructor_mappings = []

        self._get_solc_mappings(srcmap)
        self._get_solc_mappings(srcmap_constructor, constructor=True)

        super().__init__(code, creation_code, name=name)

    @staticmethod
    def get_full_contract_src_maps(ast: Dict) -> Set[str]:
        """
        Takes a solc AST and gets the src mappings for all the contracts defined in the top level of the ast
        :param ast: AST of the contract
        :return: The source maps
        """
        source_maps = set()
        for child in ast["nodes"]:
            if child.get("contractKind"):
                source_maps.add(child["src"])
        return source_maps

    def get_source_info(self, address, constructor=False):
        """

        :param address:
        :param constructor:
        :return:
        """
        disassembly = self.creation_disassembly if constructor else self.disassembly
        mappings = self.constructor_mappings if constructor else self.mappings
        index = helper.get_instruction_index(disassembly.instruction_list, address)

        solidity_file = self.solidity_files[mappings[index].solidity_file_idx]
        filename = solidity_file.filename

        offset = mappings[index].offset
        length = mappings[index].length

        code = solidity_file.data.encode("utf-8")[offset : offset + length].decode(
            "utf-8", errors="ignore"
        )
        lineno = mappings[index].lineno
        return SourceCodeInfo(filename, lineno, code, mappings[index].solc_mapping)

    def _is_autogenerated_code(self, offset: int, length: int, file_index: int) -> bool:
        """
        Checks whether the code is autogenerated or not
        :param offset: offset of the code
        :param length: length of the code
        :param file_index: file the code corresponds to
        :return: True if the code is internally generated, else false
        """

        """Handle internal compiler files
        The second condition will be placed till 
        https://github.com/ethereum/solidity/issues/10300 is dealt
        """
        if file_index == -1 or file_index >= len(self.solidity_files):
            return True
        # Handle the common code src map for the entire code.
        if (
            "{}:{}:{}".format(offset, length, file_index)
            in self.solidity_files[file_index].full_contract_src_maps
        ):
            return True

        return False

    def _get_solc_mappings(self, srcmap, constructor=False):
        """

        :param srcmap:
        :param constructor:
        """
        mappings = self.constructor_mappings if constructor else self.mappings
        prev_item = ""
        for item in srcmap:
            if item == "":
                item = prev_item
            mapping = item.split(":")

            if len(mapping) > 0 and len(mapping[0]) > 0:
                offset = int(mapping[0])

            if len(mapping) > 1 and len(mapping[1]) > 0:
                length = int(mapping[1])

            if len(mapping) > 2 and len(mapping[2]) > 0:
                idx = int(mapping[2])

            if self._is_autogenerated_code(offset, length, idx):
                lineno = None
            else:
                lineno = (
                    self.solidity_files[idx]
                    .data.encode("utf-8")[0:offset]
                    .count("\n".encode("utf-8"))
                    + 1
                )
            prev_item = item
            mappings.append(SourceMapping(idx, offset, length, lineno, item))
